<?php
/**
 * Created by PhpStorm.
 * User: EzyVA
 * Date: 3/18/2015
 * Time: 7:06 PM
 */

namespace DynamicSync;


class DynamicSync {
    private $api_key;
    private $API_URL_LOCAL = "http://apidynamicsync.dev/v1/"; //Temporary only, used on localhost
    private $API_URL_DEV = "http://sapi.dynamicsync.com/v1/";
    private $API_URL = "http://api.dynamicsync.com/v1/";
    private $mode = 0; //0 - Live, 1 - Development;
    private $events = array();
    private $eventDates = array();

    /**
     * Initializes DynamicSync and gets the list of events from the server.
     *
     * @param $api_key
     * @param $mode //0 - Live, 1 - Development;
     */
    function __construct( $api_key, $mode = 0 ) {
        $this->api_key = $api_key;
        $this->mode = $mode;
        $this->loadEvents();
        $this->loadEventDates();
    }

    /**
     * Gets the DynamicSync API key of this class' instance
     *
     * @return string
     */
    public function getAPIKey() {
        return $this->api_key;
    }

    /**
     * Sets the DynamicSync API key of this class' instance
     *
     * @param $api_key
     */
    public function setAPIKey( $api_key ) {
        $this->api_key = $api_key;
    }

    /**
     * Gets the DynamicSync API URL (Constant)
     *
     * @param $mode //0 - Live, 1 - Development;
     * @return string
     */
    public function getAPIUrl( $mode = 1 ) {
        if ( $mode == 1 )
            return $this->API_URL_DEV;
        else if ( $mode == 2 )
            return $this->API_URL_LOCAL;
        else
            return $this->API_URL;
    }

    /**
     * Sets this class' instance mode to Live or Development mode
     * @param $mode //0 - Live, 1 - Development;
     */
    public function setMode( $mode ) {
        if ( $mode == 1 )
            $this->mode = 1;
        else
            $this->mode = 0;
    }

    /**
     * Gets this class' instance mode
     * @return int //0 - Live, 1 - Development;
     */
    public function getMode() {
        return $this->mode;
    }

    /**
     * Performs cURL requests (GET, POST, PUT, DELETE)
     *
     * @param $type //GET, POST, PUT, DELETE
     * @param $method
     * @param array $args
     * @param int $timeout
     * @param bool $verify_ssl
     * @return bool|mixed
     */
    protected function sendRequest( $type, $method, $args = array(), $timeout = 30, $verify_ssl = false ) {

        $accepted_types = ['GET', 'POST', 'PUT', 'DELETE'];
        $json_data = '';

        if ( !empty($method) ) {
            if ( is_array( $method ) )
                $method = implode('/', $method);
        }


        $url = $this->getAPIUrl( $this->mode ) . $method;

        //check if type is valid
        if ( !in_array( strtoupper($type), $accepted_types ) )
            return false;

        //check if key: key is not yet on the $args array
        if ( !array_key_exists('key', $args) )
            $args['key'] = $this->api_key;


        if ( 'GET' == strtoupper($type) ) {              //Prepare URL for GET request
            $url .= '/?' . http_build_query( $args );
        }
        else if ( 'PUT' == strtoupper($type) || 'DELETE' == strtoupper($type) || 'POST' == strtoupper($type) ) {       //Prepare Data for PUT request
            $json_data = http_build_query($args);
        }
        else {                                          //Prepare Data for POST and DELETE request
            $json_data = json_encode($args);
        }

        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_HTTPHEADER, array('Content-Type: application/json'));
        curl_setopt($ch, CURLOPT_USERAGENT, 'PHP-DSAPI/1.0');
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_TIMEOUT, $timeout);

        if ( 'POST' == strtoupper($type) ) {
            curl_setopt($ch, CURLOPT_POST, true);
            curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "POST");
        }
        if ( 'PUT' == strtoupper($type) ) {
            curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "PUT");
        }
        if ( 'DELETE' == strtoupper($type) ) {
            curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "DELETE");
        }

        if ( 'GET' != strtoupper($type) ) {
            curl_setopt($ch, CURLOPT_POSTFIELDS, $json_data);
        }

        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, $verify_ssl);

        $result = curl_exec($ch);

        $http_code = curl_getinfo($ch, CURLINFO_HTTP_CODE);

        curl_close($ch);

        if ( $result )
        {
            $http_response = new HttpResponse( $http_code );
            return array(
                'http_status' => $http_response->getCode(),
                'http_message' => $http_response->getText(),
                'data' => json_decode($result, true)
            );
        }
        else
            return false;
    }

    public function requestSuccess( $result ) {
        if ( $result !== false ) {
            if ( is_array( $result ) ) {
                if ( array_key_exists( 'http_status', $result ) && '200' == $result['http_status'] )
                    return $result;
                elseif ( array_key_exists( 'http_status', $result ) && '200' != $result['http_status'] )
                    return $result;
                else
                    return false;
            }
            else
                return $result;
        }
        else
            return false;
    }
    /*
        -------------------------
          EVENTS CRUD functions
        -------------------------
    */

    /**
     * Gets the list of events from the server (JSON DECODED format)
     *
     * NOTE: This must be called at a minimum level (i.e. on Class init, or if  you need to get a refreshed event list from the server),
     *       this ensures minimal API Calls, and reduces server load due to possible multiple requests.
     *       - Event list are save into this class' property: $events. To repeatedly get this list, use the getEvents() function.
     *
     * @return bool|mixed
     */
    private function loadEvents() {
        $res = $this->sendRequest( 'GET', 'event' );
        $res = $this->requestSuccess($res);
        if ( $res !== false ) {
            $events = array();
            foreach( $res['data'] as $ev => $e ) {
                $events[] = new Event( $e['guid'], $e );
            }
            $this->events = $events;
            return true;
        }
        else
            return false;
    }

    /**
     * Gets the loaded events list
     *
     * @return array
     */
    public function getEvents() {
        $events = $this->events;
        return $events;
    }

    /**
     * Creates a new event
     *
     * @param $args
     * @return bool|mixed
     */
    public function createEvent( $args ) {

        //TODO: perform getEventTypes function here.

        $default = [
            'title' => '',
            'start_date' => '',
        ];

        $args = array_merge( $default, $args );

        if ( empty($args['title']) || empty($args['start_date'])  )
            return false;

        $res = $this->sendRequest( 'POST', 'event', $args );

        $res = $this->requestSuccess($res);
        if ( $res !== false ) {
            $this->loadEvents(); //Refresh the event's list
            return $res;
        }
        else
            return false;
    }

    /**
     * Updates the event's details
     *
     * @param $eventId
     * @param $args
     * @return bool|mixed
     */
    public function updateEvent( $args, $eventId = '' ) {

        if ( !array_key_exists('eventId', $args) && empty($eventId) )
            return false;

        if ( !array_key_exists('eventId', $args) && !empty($eventId) )
            $args['eventId'] = $eventId;

        if ( !array_key_exists('eventId', $args) && !empty($eventId) && empty($args['eventId']) )
            $args['eventId'] = $eventId;

        //Perform getEventDetails function here. This will prevent the existing event default data from being deleted/overwritten.
        $event_details = $this->getEventDetails( $args['eventId'] );

        //Get Data from the Event object
        $event_details = $event_details->getEventData();

        if ( !empty( $event_details ) )
            $args = array_merge($event_details, $args);

        $res = $this->sendRequest( 'PUT', 'event', $args );

        $res = $this->requestSuccess($res);
        if ( $res !== false ) {
            $this->loadEvents(); //Refresh the event's list
            return $res;
        }
        else
            return false;

    }

    /**
     * Deletes an event specified by event id
     *
     * @param $eventId
     * @return bool|mixed
     */
    public function deleteEvent( $eventId ) {

        $args['eventId'] = $eventId;

        $res = $this->sendRequest( 'DELETE', 'event', $args );

        $res = $this->requestSuccess($res);
        if ( $res !== false ) {
            $this->loadEvents(); //Refresh the event's list
            return $res;
        }
        else
            return false;
    }

    /**
     * Gets an event's data from the loaded events based on the given event id.
     *
     * @param $eventId
     * @return array
     */
    public function getEventDetails( $eventId ) {
        $event_details = array();

        foreach( $this->events as $events => $event ) {

            if ( $eventId == $event->getEventID() ) {
                $event_details = $event;
                break;
            }
        }

        return $event_details;
    }

    /**
     * Gets the list of event dates from the server (JSON DECODED format)
     *
     * NOTE: This must be called at a minimum level (i.e. on Class init, or if  you need to get a refreshed event date list from the server),
     *       this ensures minimal API Calls, and reduces server load due to possible multiple requests.
     *       - Event date list are save into this class' property: $eventDates. To repeatedly get this list, use the getEventDates() function.
     *
     * @return bool|mixed
     */
    private function loadEventDates() {
        $method = array('event', 'date');
        $args = array('eventId' => '');

        $res = $this->sendRequest('GET', $method, $args);
        $res = $this->requestSuccess($res);
        if ( $res !== false ) {

            $eventDates = array();
            foreach( $res['data'] as $ev => $e ) {
                $eventDates[] = new EventDate( $e['guid'], $e );
            }

            $this->eventDates = $eventDates;
            return $res;
        }
        else
            return false;
    }

    /**
     * Gets the loaded event dates of the specified event ID. If event ID is not specified, it will return all event dates for all events
     * @param string $eventID
     * @param string $sortBy
     * @param array $filter
     * @param int $order
     * @return array
     */
    public function getEventDates( $eventID = '', $filter = array(), $sortBy = 'start_date', $order = SORT_ASC)
    {
        $event_dates = array();

        //Get Array data of each Event Date Class
        foreach( $this->eventDates as $ed => $e ) {
            $event_dates[] = $e->getEventDateData();
        }

        $filteredDates = array();

        if ( empty( $eventID ) ) {
            $filteredDates = $event_dates;
        }
        else {
            foreach( $event_dates as $e_dates => $event_date ) {
                if ( $eventID == $event_date['eventId'] ) {
                    $filteredDates[] = $event_date;
                }
            }
        }

        if ( !empty($filteredDates) ) {
            $this->array_sort_by_column($filteredDates, $sortBy, $order);
            //usort($events, $this->make_comparer($sortBy));
        }

        if ( !empty($filter) && !empty($filteredDates) )
            $filteredDates = $this->applyDataFilters($filteredDates, $filter);

        //Data has been sorted/filtered, convert them back to Event Date object
        $eventDatesObject = array();

        foreach( $filteredDates as $fd => $eventDate ) {
            $eventDatesObject[] = new EventDate( $eventDate['guid'], $eventDate );
        }

        return $eventDatesObject;
    }

    /**
     * @param $eventDateId
     * @return bool|mixed
     */
    public function getEventDate( $eventDateId ) {
        if ( empty($eventDateId) )
            return false;

        $event_dates = $this->eventDates;

        foreach( $event_dates as $e_dates => $event_date ) {
            if ( $eventDateId == $event_date->getEventDateID() ) {
                return $event_date;
            }
        }

        return false;
    }

    /**
     * Creates new Event Date for the specified Event ID
     *
     * @param $eventId
     * @param $args
     * @return bool|mixed
     */
    public function createEventDate( $args, $eventId = '' ) {

        if ( empty( $args ) )
            return false;

        //required field not specified
        if ( !array_key_exists('start_date', $args) )
            return false;

        //check if date is valid: YYYY-MM-DD hh:mm:ss
        $date = date_parse($args['start_date']);
        if ( $date["error_count"] != 0 || !checkdate($date["month"], $date["day"], $date["year"]) )
            return false;

        //no date specified, allow API to allow allocate the end date
        if ( !array_key_exists('end_date', $args) )
            $args['end_date'] = '';

        if ( !array_key_exists('eventId', $args) && empty($eventId) )
            return false;

        if ( !array_key_exists('eventId', $args) && !empty($eventId) )
            $args['eventId'] = $eventId;

        if ( !array_key_exists('eventId', $args) && !empty($eventId) && empty($args['eventId']) )
            $args['eventId'] = $eventId;

        //Get Default Data from the parent Event
        $event_details = $this->getEventDetails( $args['eventId'] );
        $event_details = $event_details->getEventData();

        if ( array_key_exists('timezone', $args) ) {
            if ( empty($args['timezone']) && $args['eventId'] != '0' ) {
                //timezone is '', setup default timezone
                $args['timezone'] = $event_details['timezone'];
            }
        }
        else {
            $args['timezone'] = $event_details['timezone'];
        }

        $method = array('event', 'date');

        $res = $this->sendRequest( 'POST', $method, $args );

        $res = $this->requestSuccess($res);
        if ( $res !== false ) {
            $this->loadEventDates(); //Refresh the event's list
            return $res;
        }
        else
            return false;
    }

    public function updateEventDate( $args, $eventDateID = '' ) {

        if ( empty( $args ) )
            return false;

        //required field not specified
        if ( !array_key_exists('start_date', $args) )
            return false;

        //check if date is valid: YYYY-MM-DD hh:mm:ss
        $date = date_parse($args['start_date']);
        if ( $date["error_count"] != 0 || !checkdate($date["month"], $date["day"], $date["year"]) )
            return false;

        //no date specified, allow API to allow allocate the end date
        if ( !array_key_exists('end_date', $args) )
            $args['end_date'] = date('Y-m-d H:i:s', strtotime($args['start_date'].' + 1 hour')); // 3600 sec. = 1 hour

        if ( !array_key_exists('eventDateId', $args) && empty($eventDateID) )
            return false;

        if ( !array_key_exists('eventDateId', $args) && !empty($eventDateID) )
            $args['eventDateId'] = $eventDateID;

        if ( !array_key_exists('eventDateId', $args) && !empty($eventDateID) && empty($args['eventDateId']) )
            $args['eventDateId'] = $eventDateID;

        $whitelist = array(
            'eventId',
            'eventDateId',
            'start_date',
            'end_date',
            'timezone',
            'WebinarJoinUrl',
            'title',
            'description',
            'pretty_time',
            'sms_reminder_time',
            'venue_name',
            'venue_address',
            'contact_person',
            'contact_number',
            'price',
            'max_attendees',
            'user1',
            'user2',
        );

        $this->write_log( 'BEFORE getEventData()' . print_r($args, true) );

        /*
         *  EVENT DATE CLASS OBJECT - Get current data of this Event Date
         */
        //Perform getEventDateDetails function here. This will prevent the existing event date default data from being deleted/overwritten.
        $event_date_details = $this->getEventDateDetails( $args['eventDateId'] );
        $event_date_details = $event_date_details->getEventDateData();

        //Merge this event date's current data to the new data.
        if ( !empty( $event_date_details ) ) {
            $filtered = array_intersect_key( $event_date_details, array_flip( $whitelist ) );
            $this->write_log( 'FILTER 1: ' . print_r($filtered, true) );

            $args = array_merge($filtered, $args);

            $this->write_log( 'MERGED 1: ' . print_r($args, true) );
        }



        /*
         * EVENT CLASS OBJECT - Get current data of this Event
         */
        //Perform getEventDetails function here. This will prevent the existing event default data from being deleted/overwritten.
        $event_details = $this->getEventDetails( $args['eventId'] );
        $event_details = $event_details->getEventData();

        //If there are other default data not specified on the update data, merge it to the parent event's default data.
        if ( !empty( $event_details ) ) {
            $filtered = array_intersect_key( $event_details, array_flip( $whitelist ) );
            $this->write_log( 'FILTER 2: ' . print_r($filtered, true) );

            $args = array_merge($filtered, $args);

            $this->write_log( 'MERGED 2: ' . print_r($args, true) );
        }

        $this->write_log( 'AFTER getEventData()' . print_r($args, true) );

        $method = array('event', 'date');
        $res = $this->sendRequest( 'PUT', $method, $args );

        $res = $this->requestSuccess($res);
        if ( $res !== false ) {
            $this->loadEventDates(); //Refresh the event dates list
            return $res;
        }
        else
            return false;
    }

    public function getEventDateDetails( $eventDateID ) {
        $event_date_details = array();

        foreach( $this->eventDates as $ed => $eventDate ) {
            if ( $eventDateID == $eventDate->getEventDateID() ) {
                $event_date_details = $eventDate;
                break;
            }
        }

        return $event_date_details;
    }


    /**
     * Gets list of attendee based on the specified Event ID or Event Date ID
     * If Event ID is specified, gets all attendees from all event dates
     * If Event Date ID is specified, gets all attendees from the event date id
     * If both Event ID and Event Date ID is specified, the Event Date ID will take precedence and will be used.
     *
     * @param $args
     * @param $eventID
     * @param $eventDateID
     * @return bool|mixed
     */
    public function getAttendees( $args = array(), $eventID = '', $eventDateID = '', $filter = array() ) {

        if ( !is_array( $args ) ) {
            $args = array();
        }

        //check if both eventId and eventDateId is not specified
        if ( !array_key_exists('eventId', $args) && empty($eventID) && !array_key_exists('eventDateId', $args) && empty($eventDateID) )
            return false;

        If ( !array_key_exists('eventId', $args) && !empty($eventID) )
            $args['eventId'] = $eventID;

        If ( !array_key_exists('eventDateId', $args) && !empty($eventDateID) ) {
            $args['eventDateId'] = $eventDateID;

            //eventDateId has been specified, remove eventId
            if ( array_key_exists('eventId', $args) )
                unset($args['eventId']);
        }

        $method = array('event', 'attendee');

        $res = $this->sendRequest('GET', $method, $args);

        $res = $this->requestSuccess($res);
        if ( $res !== false ) {

            $filteredAttendees = $res['data'];

            if ( !empty($filter) && is_array($filter) ) {
                $filteredAttendees = $this->applyDataFilters( $filteredAttendees, $filter );
            }

            //Attendees has been sorted/filtered, convert them back to Attendee object
            $attendeesObject = array();

            foreach( $filteredAttendees as $fa => $attendee ) {
                $attendeesObject[] = new Attendee( $attendee['guid'], $attendee );
            }

            return $attendeesObject;

        }
        else
            return false;
    }

    /**
     * Creates an attendee on an specified Event ID or Event Date ID
     * @param $args
     * @return bool|mixed
     */
    public function createAttendee( $args ) {

        if ( !array_key_exists('contactId', $args) ) {
            if ( !array_key_exists('contactEmail', $args) )
                return false;
        }

        if ( empty($args['contactId']) ) {
            if ( empty($args['contactEmail']) )
                return false;
        }

        if ( !array_key_exists('eventId', $args) && !array_key_exists('eventDateId', $args) )
            return false;

        if ( empty($args['eventId']) && empty($args['eventDateId']) )
            return false;

        if ( !empty($args['eventId']) && !empty($args['eventDateId']) )
            unset($args['eventId']); //event date id has been specified, remove event id

        if ( empty($args['eventId']) && !empty($args['eventDateId']) )
            unset($args['eventId']); //event date id has been specified, remove event id

        if ( !empty($args['eventId']) && empty($args['eventDateId']))
            unset($args['eventDateId']); //event date id has not been specified, default to event id, and remove event date id

        $method = array('event', 'attendee');

        $res = $this->sendRequest('POST', $method, $args);

        $res = $this->requestSuccess($res);
        if ( $res !== false ) {
            return $res;
        }
        else
            return false;

    }

    /**
     * Updates an attendee based on the specified Attendee ID
     *
     * @param $args
     * @param string $attendeeID
     * @return bool
     */
    public function updateAttendee( $args, $attendeeID = '' ) {

        if ( !array_key_exists('attendeeId', $args) && empty($attendeeID) )
            return false;

        if ( !array_key_exists('contactEmail', $args) )
            return false;

        if ( empty($args['contactEmail']) )
            return false;

        //override attendeeId on $args if specified on function parameter
        if ( !array_key_exists('attendeeId', $args) && !empty($attendeeID) )
            $args['attendeeId'] = $attendeeID;

        if ( !array_key_exists('updateInfusionsoftContact', $args) )
            $args['updateInfusionsoftContact'] = 'on'; //Set default to update Infusionsoft contact details

        $method = array('event', 'attendee');

        $res = $this->sendRequest('PUT', $method, $args);

        $res = $this->requestSuccess($res);
        if ( $res !== false ) {
            return $res;
        }
        else
            return false;

    }

    protected function write_log( $msg )
    {
        $file = dirname(__FILE__).'/ds_log.log';

        $date = date('m/d/Y h:i:s a');

        $msg = '[ '.$date.' ]: '.$msg;
        $msg .= "\r\n\r\n";
        // The new person to add to the file
        // Write the contents to the file,
        // using the FILE_APPEND flag to append the content to the end of the file
        // and the LOCK_EX flag to prevent anyone else writing to the file at the same time
        file_put_contents($file, $msg, FILE_APPEND | LOCK_EX);
    }

    protected function array_sort_by_column(&$arr, $col, $dir = SORT_ASC) {
        $sort_col = array();
        foreach ($arr as $key=> $row) {
            $sort_col[$key] = $row[$col];
        }
        array_multisort($sort_col, $dir, $arr);
    }

    /**
     * Apply data filtering based on the passed filter
     *
     *
     * $filter = array(
     *     'key' => 'column name to perform comparison',
     *     'compare' => 'comparison operator, see below for accepted operators',
     *     'value' => 'value to compare'
     * );
     *
     * compare key accepts: contains, ==, !=, >=, <=, >, <
     *
     * @param $data
     * @param array $filter
     * @return array|bool
     */
    protected function applyDataFilters( $data, $filter = array() ) {
        if ( empty($filter) || empty($data) )
            return false;

        if ( !array_key_exists('key', $filter) || !array_key_exists('compare', $filter) || !array_key_exists('value', $filter) )
            return false;

        $filteredData = array();

        foreach( $data as $dat => $d ) {
            if ( array_key_exists($filter['key'], $d) ) {
                switch( $filter['compare'] ) {
                    case "contains":
                        $res = strpos($d[ $filter['key'] ], $filter['value'] );
                        if (  $res !== false && $res >= 0 )
                            $filteredData[] = $d;
                        break;

                    case "==";
                        if ( $d[ $filter['key'] ] == $filter['value'] )
                            $filteredData[] = $d;
                        break;

                    case "!=";
                        if ( $d[ $filter['key'] ] != $filter['value'] )
                            $filteredData[] = $d;
                        break;

                    case ">=";
                        if ( $d[ $filter['key'] ] >= $filter['value'] )
                            $filteredData[] = $d;
                        break;

                    case "<=";
                        if ( $d[ $filter['key'] ] <= $filter['value'] )
                            $filteredData[] = $d;
                        break;

                    case ">";
                        if ( $d[ $filter['key'] ] > $filter['value'] )
                            $filteredData[] = $d;
                        break;

                    case "<";
                        if ( $d[ $filter['key'] ] < $filter['value'] )
                            $filteredData[] = $d;
                        break;
                }
            }
        }

        return $filteredData;
    }

    protected function make_comparer() {
        // Normalize criteria up front so that the comparer finds everything tidy
        $criteria = func_get_args();
        foreach ($criteria as $index => $criterion) {
            $criteria[$index] = is_array($criterion)
                ? array_pad($criterion, 3, null)
                : array($criterion, SORT_ASC, null);
        }

        return function($first, $second) use (&$criteria) {
            foreach ($criteria as $criterion) {
                // How will we compare this round?
                list($column, $sortOrder, $projection) = $criterion;
                $sortOrder = $sortOrder === SORT_DESC ? -1 : 1;

                // If a projection was defined project the values now
                if ($projection) {
                    $lhs = call_user_func($projection, $first[$column]);
                    $rhs = call_user_func($projection, $second[$column]);
                }
                else {
                    $lhs = $first[$column];
                    $rhs = $second[$column];
                }

                // Do the actual comparison; do not return if equal
                if ($lhs < $rhs) {
                    return -1 * $sortOrder;
                }
                else if ($lhs > $rhs) {
                    return 1 * $sortOrder;
                }
            }

            return 0; // tiebreakers exhausted, so $first == $second
        };
    }

}